'
' -------------------------------------------------------------------------------------------------------------------
'
'	This file is part of EliteGameEngine source distribution
'
' -------------------------------------------------------------------------------------------------------------------
'
'{
'{	#defines:
'
'		ENGINE_ENABLE_FPS
'			- #define this symbol to enable FPS counting functions
'}
'
'
'{	Enumerations:
'
'
'		Enum TransitionEffect
'			None
'			Fade
'			SlideIn
'			SlideOut
'			ZoomIn
'			ZoomOut
'			HorisontalStripes
'			VerticalStripes
'			Random
'		End Enum
'			
'			- game screen transition effects
'
'
'		Enum TransitionDirection
'			FromLeft
'			FromRight
'			FromTop
'			FromBottom
'			FromTopLeft
'			FromTopRight
'			FromBottomLeft
'			FromBottomRight
'		End Enum
'
'			- direction of transition effect (used with SlideIn and SlideOut effects)
'
'
'		Enum InputEvent
'			None
'			KeyPress
'			KeyRelease
'			KeyRepeat
'			MouseMove
'			MouseButtonPress
'			MouseButtonRelease
'			MouseButtonDoubleClick
'			MouseWheel
'			MouseHWheel
'		End Enum
'
'			- input events available in onFrame handler
' 
'}
'
'
'{	Structures:
'
'
'		Type Resolution
'			Width	As Integer
'			Height	As Integer
'			Operator Cast As String
'		End Type
'
'			- holds information about screen resolution
'
'
'		Type ResolutionList
'			List	As Engine.Resolution Pointer
'			Count	As Integer
'			Sub Add(w As Integer, h As Integer)
'		End Type
'
'			- list of screen resolutions
'
'
'		Type Region
'			Left	As Integer
'			Right	As Integer
'			Width	As Integer
'			Height	As Integer
'			Dirty	As Elite.Boolean
'		End Type
'
'			- represents screen region
'
'
'		Type GameScreenHandle As Integer
'			- game screen handle
'
'
'		
'		Type GameScreenEvents
'
'			Event onPaint(dest As Any Pointer)
'				- raised when entire screen needs to be repainted (dest is destination image handle)
'
'			Event onPaintRegion(dest As Any Pointer, rgn As Engine.Region)
'				- raised when specified region needs to be repainted
'
'			Event onShow
'				- raised when screen is shown
'
'			Event onHide(newScreen As Engine.GameScreenHandle, ByRef Cancel As Elite.Boolean = Elite.False)
'				- raised when screen is about to be hidden
'				  newScreen is handle to the screen which is about to be shown, this parametar is set
'				  to Engine.IllegalScreenHandle when exiting from application
'
'			Event onFrame(fElapsed As Single, eInputEvent As InputEvent, eventData As FB.EVENT)
'				- raised on every frame 
'				  fElapsed is number of seconds passed since last frame
'				  eInputData is set to one of InputEvents 
'				  eventData holds event specific data
'
'
'		End Type
'
'			- game screen events
'}
'
'
'{	Variables:
'
'		Initialized As Elite.Boolean = Elite.False
'			- this is set to Elite.True once the Initialize is called
'
'		GameScreen() As GameScreenEvents
'			- array of game screens
'
'		ScreenCount As Integer
'			- number of game screens
'
'		ActiveGameScreen As GameScreenHandle
'			- currently displaying screen
'}
'
'
'{	Events:
'
'		onInitialize()
'			- raised when engine is ready and game code can start executing
'
'		onShutdown(ByRef Cancel As Elite.Boolean)
'			- raised when Engine.Shutdown is called (set Cancel to Elite.True in handler, to cancel shutdown)
'
'		onDestroy()
'			- raised when exiting from application
'
'		onStart()
'			- raised when game is started (just before game loop starts) 
'
'		onMouseEnter()
'			- raised when mouse enters window's area
'
'		onMouseLeave()
'			- raised when mouse leaves window's area
'
'		onFocus()
'			- raised when window gets focus
'
'		onBlur()
'			- raised when window loses focus
'
'}
'
'
'{	Methods:
'
'
'		Function Initialize(res As Engine.Resolution, fs As Elite.Boolean = Elite.False, flags As Integer = 0) As Elite.Boolean
'		Function Initialize(w As Integer, h As Integer, fs As Elite.Boolean = Elite.False, flags As Integer = 0) As Elite.Boolean
'			- initializes engine to specified window mode
'			  res - Engine.Resolution value
'			  w - width (in pixels)
'			  h - height (in pixels)
'			  fs - windowed (False) or fullscreen (True) mode
'			  flags - additional ScreenRes flags
'			  returns Elite.False if failed to set video mode
'
'
'		Function NewGameScreen As GameScreenHandle
'			- creates new game screen, returns screen's handle
'
'		Sub Show(scr As GameScreenHandle, transMethod As Engine.TransitionEffect = Engine.TransitionEffect.None, transDir As Engine.TransitionDirection = Engine.TransitionDirection.FromLeft)
'			- shows game screen and sets that screen as active (performs transition effect if one is chosen)
'
'
'		Function Shutdown(Force As Elite.Boolean = Elite.False) As Elite.Boolean
'			- call this if you want to exit from application (game)
'			  if you set Force to Elite.True then handlers will not be allowed to cancel shutdown process
'			  returns Elite.False if shutdown was canceled 
'
'
'		Function SupportedResolutions As Engine.ResolutionList 
'			- returns available fullscreen video modes (modes are returned from lowest to highest supported)
'}
'}

#Ifndef __EGE_ENGINE__
#Define __EGE_ENGINE__
	
	Namespace Engine
		
		'{ declare enumerations
		
		' game screen transition effects
		Enum TransitionEffect
			None
			Fade
			SlideIn
			SlideOut
			ZoomIn
			ZoomOut
			HorisontalStripes
			VerticalStripes
		End Enum
		
		' direction of transition effect (used with SlideIn and SlideOut effects)
		Enum TransitionDirection
			FromLeft
			FromRight
			FromTop
			FromBottom
			FromTopLeft
			FromTopRight
			FromBottomLeft
			FromBottomRight
		End Enum

		' input events available in onFrame handler
		Enum InputEvent
			None						= -1
			KeyPress					= FB.EVENT_KEY_PRESS
			KeyRelease					= FB.EVENT_KEY_RELEASE
			KeyRepeat					= FB.EVENT_KEY_REPEAT
			MouseMove					= FB.EVENT_MOUSE_MOVE
			MouseButtonPress			= FB.EVENT_MOUSE_BUTTON_PRESS
			MouseButtonRelease			= FB.EVENT_MOUSE_BUTTON_RELEASE
			MouseButtonDoubleClick		= FB.EVENT_MOUSE_DOUBLE_CLICK
			MouseWheel					= FB.EVENT_MOUSE_WHEEL
			MouseHWheel					= FB.EVENT_MOUSE_HWHEEL
		End Enum

		'}
		
		'{ declare types
		
		' holds information about screen resolution
		Type Resolution
			
			Width	As Integer
			Height	As Integer
			
			Declare Operator Cast As String
			
		End Type
		
		' list of screen resolutions
		Type ResolutionList
			
			Count	As Integer
			List	As Engine.Resolution Pointer
			
			Declare Sub Add(w As Integer, h As Integer)
			
		End Type
		
		' represents screen region
		Type Region
			Left	As Integer
			Right	As Integer
			Width	As Integer
			Height	As Integer
			Dirty	As Elite.Boolean
		End Type
		
		' game screen handle
		Type GameScreenHandle As Integer
		
		Const IllegalScreenHandle	As Engine.GameScreenHandle = -1
		
		
		SimpleEvent( onShow )
		CreateEvent( onHide, (newScreen, Cancel), (newScreen As Engine.GameScreenHandle, ByRef Cancel As Elite.Boolean = Elite.False) )
		CreateEvent( onPaint, (dest), (dest As Any Pointer) )
		CreateEvent( onPaintRegion, (dest, rgn), (dest As Any Pointer, rgn As Engine.Region) )
		CreateEvent( onFrame, (fElapsed, eInputEvent, eventData), (fElapsed As Single, eInputEvent As InputEvent, eventData As FB.EVENT) )
		
		' game screen events
		Type GameScreenEvents
			onShow			As onShowEvent
			onHide			As onHideEvent
			onPaint			As onPaintEvent
			onPaintRegion	As onPaintRegionEvent
			onFrame			As onFrameEvent
		End Type
		
		'}
		
		
		'{ declare variables
		
		Dim Shared As Elite.Boolean			Initialized = Elite.False				' current engine state
		Dim Shared As GameScreenEvents		GameScreen()							' array of game screens
		Dim Shared As Integer				ActiveGameScreen = IllegalScreenHandle	' currently displaying screen
		Dim Shared As Integer				ScreenCount = 0							' number of game screens
		
		Dim Shared As Elite.Boolean			GameLoopRunning = Elite.False			' is the game loop running?
		Dim Shared As Elite.Boolean			Started = Elite.False					' are we Started already?

		'}
		

		'{ declare events
		
		SimpleEvent( onInitialize )
		CreateEvent( onShutdown, (Cancel), (ByRef Cancel As Elite.Boolean) )
		SimpleEvent( onDestroy )
		SimpleEvent( onStart )
		SimpleEvent( onMouseEnter )
		SimpleEvent( onMouseLeave )
		SimpleEvent( onFocus )
		SimpleEvent( onBlur )
		
		Dim Shared onInitialize	As onInitializeEvent	' engine is ready and game code can start executing
		Dim Shared onShutdown	As onShutdownEvent		' Engine.Shutdown is called (set Cancel to Elite.True in handler, to cancel shutdown)
		Dim Shared onDestroy	As onDestroyEvent		' exiting from application
		Dim Shared onStart		As onStartEvent			' game is started (just before game loop starts)
		Dim Shared onMouseEnter	As onMouseEnterEvent	' mouse enters window's area
		Dim Shared onMouseLeave	As onMouseLeaveEvent	' mouse leaves window's area
		Dim Shared onFocus		As onFocusEvent			' window gots focus
		Dim Shared onBlur		As onBlurEvent			' window loses focus

		'}


		'{ declare methods
		
		Declare Function Initialize OverLoad(res As Engine.Resolution, fs As Elite.Boolean = Elite.False, flags As Integer = 0) As Elite.Boolean
		Declare Function Initialize (w As Integer, h As Integer, fs As Elite.Boolean = Elite.False, flags As Integer = 0) As Elite.Boolean
		
		Declare Function NewGameScreen As GameScreenHandle 
		Declare Sub Show(scr As GameScreenHandle, transMethod As Engine.TransitionEffect = Engine.TransitionEffect.None, transDir As Engine.TransitionDirection = Engine.TransitionDirection.FromLeft)
		
		Declare Sub Start
		
		Declare Function Shutdown(Force As Elite.Boolean = Elite.False) As Elite.Boolean
		
		Declare Function SupportedResolutions As Engine.ResolutionList
		
		'}
		
		
		'{ implement types
		
		Operator Resolution.Cast As String
			
			Return Str(This.Width) & "x" & Str(This.Height)
			
		End Operator
		
		
		Sub ResolutionList.Add(w As Integer, h As Integer)
			
			This.Count += 1
			
			This.List = ReAllocate(This.List, This.Count * SizeOf(Engine.Resolution))
			
			This.List[This.Count - 1] = Type<Engine.Resolution>(w, h)
			
		End Sub
		
		'}
		
		
		'{ implement methods
		
		' initializes engine to specified window mode
		'	res - Engine.Resolution value
		'	w - width (in pixels)
		'	h - height (in pixels)
		'	fs - windowed (False) or fullscreen (True) mode
		'	flags - additional ScreenRes flags
		'	returns Elite.False if failed to set video mode		
		Function Initialize(res As Engine.Resolution, fs As Elite.Boolean = Elite.False, flags As Integer = 0) As Elite.Boolean
			
			Return Engine.Initialize(res.Width, res.Height, fs, flags)

		End Function
	
		Function Initialize(w As Integer, h As Integer, fs As Elite.Boolean = Elite.False, flags As Integer = 0) As Elite.Boolean
			
			If Not Engine.Initialized Then

				ScreenRes w, h, 32, 1, IIf(fs, FB.GFX_FULLSCREEN, FB.GFX_WINDOWED) Or flags
				
				If ScreenPtr = 0 Then

					Return Elite.False			' Failed to set video mode

				EndIf
				
				SetMouse w / 2, h / 2, 0		' center mouse in screen and hide it
				
				Randomize Timer					' seen randomize generator

				Engine.Initialized = Elite.True	' set the initialized flag

				Engine.onInitialize.Raise 		' everything ready, lets fireup the game!

			EndIf

			Return Elite.True

		End Function
		
		
		' creates new game screen, returns screen's handle
		Function NewGameScreen As GameScreenHandle
			
			ReDim Preserve Engine.GameScreen(0 To Engine.ScreenCount)
			
			Function = Engine.ScreenCount
			
			Engine.ScreenCount += 1
			
		End Function
		
		
		' shows game screen and sets that screen as active (performs transition effect if one is chosen)
		Sub Show(scr As GameScreenHandle, transMethod As Engine.TransitionEffect = Engine.TransitionEffect.None, transDir As Engine.TransitionDirection = Engine.TransitionDirection.FromLeft)
			
			If Engine.ActiveGameScreen <> scr Then
				
				If transMethod <> TransitionEffect.None Then
				
				EndIf
				
				Dim As Elite.Boolean Cancel = Elite.False
				
				Engine.GameScreen(scr).onHide.Raise scr, Cancel
				
				If Cancel Then Exit Sub
				
			EndIf
			
			Dim As Integer firstScreen = Engine.ActiveGameScreen = Engine.IllegalScreenHandle			
			
			Engine.ActiveGameScreen = scr
			
			Engine.GameScreen(scr).onShow.Raise
			
			If firstScreen Then Engine.Start
			
		End Sub
		
		
		' starts game loop
		Sub Start
			
			If Engine.Started Then
				Exit Sub	' we're already running!
			EndIf
			
			If Not Engine.ActiveGameScreen > Engine.IllegalScreenHandle Then
				Exit Sub 	' no game screen active!
			EndIf
			
			Dim As Single				fElapsed = Timer
			Dim As FB.EVENT				systemEvent
			Dim As Engine.InputEvent	eInputEvent
			
			Engine.Started = Elite.True		' we are about to start running the game!
			
			Engine.onStart.Raise			' raise onStart event
			
			GameLoopRunning = Elite.True	' set the running flag
			
			ScreenLock
			
			Do
				fElapsed = Timer - fElapsed				' how much miliseconds passed since last frame
				
				' process events
				If ScreenEvent(VarPtr(systemEvent)) Then
					
					If Not (systemEvent.Type < InputEvent.MouseButtonDoubleClick Or systemEvent.Type = InputEvent.MouseHWHeel) Then
					
						' "system" event (not an input event)
						Select Case systemEvent.Type
							
							' mouse enters window's area
							Case FB.EVENT_MOUSE_ENTER: Engine.onMouseEnter.Raise
							
							' mouse leaves window's area
							Case FB.EVENT_MOUSE_EXIT: Engine.onMouseLeave.Raise
							
							' window gets focus
							Case FB.EVENT_WINDOW_GOT_FOCUS: Engine.onFocus.Raise
								
							' window loses focus
							Case FB.EVENT_WINDOW_LOST_FOCUS: Engine.onBlur.Raise
								
							' user attempts to close window
							Case FB.EVENT_WINDOW_CLOSE: Engine.Shutdown
							
						End Select
						
						eInputEvent = Engine.InputEvent.None	' not an input event
					
					Else
					
						eInputEvent = systemEvent.Type ' keyboard or mouse event
					
					EndIf
					 
				Else
					
					eInputEvent = Engine.InputEvent.None ' no event
					
				EndIf
				
				
				' do game logic
				Engine.GameScreen(Engine.ActiveGameScreen).onFrame.Raise fElapsed, eInputEvent, systemEvent
				
				
				' redraw if needed
				
				
				
				ScreenUnLock
				
					Sleep 1, 1	' keep CPU usage at minimum level
				
				ScreenLock
				
			Loop While GameLoopRunning
			
			ScreenUnlock
			
		End Sub
		
		
		' call this if you want to exit from application (game)
		' if you set Force to Elite.True then handlers will not be allowed to cancel shutdown process
		' returns Elite.False if shutdown was canceled 
		Function Shutdown(Force As Elite.Boolean = Elite.False) As Elite.Boolean
			
			Dim As Elite.Boolean Cancel = Elite.False
			
			Engine.onShutdown.Raise Cancel
			
			If Not Force And Cancel Then
				
				Return Elite.False			' shutdown canceled
				
			EndIf
			
			' raise active screen's onHide event
			If Engine.ActiveGameScreen > Engine.IllegalScreenHandle Then
				
				Engine.GameScreen(Engine.ActiveGameScreen).onHide.Raise Engine.IllegalScreenHandle
					
			EndIf
			
			Engine.onDestroy.Raise					' raise onDestroy
			
			Engine.GameLoopRunning = Elite.False	' stops game loop
			
			Return Elite.True						' engine is destroyed now
				
		End Function
		
		
		' returns available fullscreen video modes (modes are returned from lowest to highest supported)
		Function SupportedResolutions As Engine.ResolutionList
			
			Dim As Integer 					mode
			Dim As Engine.ResolutionList 	list
			
			mode = ScreenList(32)			' get all supported resolutions for 32bit color depth
			
			While mode
				
				list.Add HiWord(mode), LoWord(mode)
				
				mode = ScreenList
				
			Wend
			
			Return list
			
		End Function
		
		'}
		
	End Namespace

#EndIf